package org.qatools.validator;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import org.apache.log4j.Logger;
import org.qatools.local.LocalDir;
import org.qatools.local.LocalDirUtil;
import org.qatools.local.LocalFile;
import org.qatools.local.LocalFileUtil;
import org.qatools.util.LongUtil;



/**
 * 
 * @author hockeyeh
 *
 */
public class DetailReport {
    private static Logger logger = Logger.getLogger(DetailReport.class);
    public static boolean logInfo = false;
    private static String ASSERTION_WITHOUT_FOLLUP_CHECK_MSG = "WARNING - an assertion was invoked without a followup junitAssertNoErrors() check";
    
	private static List<Package> packages = new LinkedList<Package>();
	private static List<ErrorDetail> errorDetails = new LinkedList<ErrorDetail>();
	private static LocalDir qatoolsOrgDir = new LocalDir("target/QAToolsOrg");
	private static LocalFile notEfficientFile = new LocalFile(qatoolsOrgDir, "not-efficient.html");
	private static Boolean notEfficientWarning = null;
	private static String parametizedValues = null;		//report unique test case names if you use parametized test cases
	
	//
	// detail reports
	private static long numDetailWrites = 0;
	private static long numErrorWrites = 0;
	private static long maxDetailWrites = -1;
	private static Boolean detailReportFlag = Boolean.TRUE;
	private static Boolean autoGenerateDetailReportFlag = Boolean.TRUE;	//Validator.junitAssertNoErrors by default writes results to disk.
																//This implementation is not efficient because there are extra file I/O																//Better implementation is if the Client invokes the generateReport when the client is done

	//
	// color codes
	// http://www.w3schools.com/html/html_colors.asp
	private static String GREEN = "00FF66";
	private static String RED = "FF6666";
	private static String YELLOW = "yellow";
	
	static {
		logger.debug("\n*****\n***** initializing DetailReport.java \n*****");
		if (isAutoGenerateDetailReportEnabled() || isDetailReportEnabled()) {
			LocalDirUtil.createDir(qatoolsOrgDir);
			createNotEfficientFile();		
		}
	}	//static
	

    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * 
	 */
	private static void createNotEfficientFile() {
		try {
			List<String> lines = DetailReport.getPageHeader("Not Efficient!");
			lines.add("The default report generator is not efficient. It does extra file I/O each time the validator.junitAssertNoErrors() is invoked because the APIs have no way of knowing when your test automation completes.");
			lines.add("You can force the reporting I/O to be more efficient with fewer I/O by doing the following:");
			lines.add("<ul>");
			lines.add("<li>Junit solution:");
			lines.add("<li>" + DetailReport.class.getSimpleName() + ".disableAutoGenerateDetailReport(); - this disables the default reporting - put it in your system initialization");
			lines.add("<li>" + DetailReport.class.getSimpleName() + ".generateDetailReport();  - do this in your @AfterClass teardown method");
			lines.add("</ul>");
			lines.add("<ul>");
			lines.add("<li>TestNG solution:");
			lines.add("<li>" + DetailReport.class.getSimpleName() + ".disableAutoGenerateDetailReport(); - this disables the default reporting - put it in your system initialization");
			lines.add("<li>" + DetailReport.class.getSimpleName() + ".generateDetailReport();  - do this in your @AfterClass teardown method");
			lines.add("</ul>");
			lines.add("To disable this message invoke this DetailReport.disableNotEfficientWarning() in your setUp");
			lines.addAll(DetailReport.getPageFooter());
			LocalFileUtil.writeLines(notEfficientFile, lines);
		} catch (Exception e) {
			logger.warn("createNotEfficientFile - failed");
		}			
		
	}
	
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * If you are using parameterized test cases, this allows you to 
	 * add the parameters to the DetailReport.
	 * 
	 * @return
	 */
	public static String getParametizedValues() {
		return parametizedValues;
	}

	/**
	 * If you are using parameterized test cases, this allows you to 
	 * add the parameters to the DetailReport.
	 * 
	 * @return
	 */
	public static void setParametizedValues(String parametizedValues) {
		DetailReport.parametizedValues = parametizedValues;
	}

	
    // ------------------------------------------------------
    // ----- enabling detail / error / autoGenerate report flags
    // ------------------------------------------------------
		

	/**
	 * Used for generating detailed reports
	 * @return
	 */
	public static Boolean isDetailReportEnabled() {
		return detailReportFlag;
	}
	
	/**
	 * Enable detail report generation
	 */
	public static void enableDetailReport() {
		detailReportFlag = Boolean.TRUE;
		logger.trace ("enableDetailReport - " + detailReportFlag);
	}
	
	/**
	 * Disable detail report generation
	 */
	public static void disableDetailReport() {
		detailReportFlag = Boolean.FALSE;
        logger.trace ("disableDetailReport - " + detailReportFlag);
	}
	
	
		
	/**
	 * QATools writes detail & error summary reports by default. It performs 
	 * extra disk I/O therefore is not as efficient as the client
	 * invoking these when done
	 * 
	 * 	DetailReport.generateDetailReport()
	 * 	DetailReport.generateErrorSummaryReport()
	 * 
	 * @return
	 */
	public static Boolean isAutoGenerateDetailReportEnabled() {
		return autoGenerateDetailReportFlag;
	}
	
	/**
	 * Enable default report generation
	 */
	public static void enableAutoGenerateDetailReport() {
		autoGenerateDetailReportFlag = Boolean.TRUE;
	}
	
	/**
	 * Disable default report generation
	 */
	public static void disableAutoGenerateDetailReport() {
		autoGenerateDetailReportFlag = Boolean.FALSE;
	}
	

	/**
	 * This allows the client to disable detail reporting after N tries. 
	 * If set then detail reporting turns itself off after N writes.
	 * May come in handy if you don't want your build systems to generate these reports 
	 * yet you want enabled for running a handful on your PC
	 * 
	 * @param maxWrites
	 */
	public static void setMaxDetailWrites(int maxWrites) {
	    maxDetailWrites = maxWrites;
	}


	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------

	/**
	 * Disable the "Not Efficient" warning that shows up on the detail report page
	 */
	public static void disableNotEfficientWarning() {
		notEfficientWarning = Boolean.FALSE;
	}
	
	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------
	
	/**
	 * Add debug info to the testcase report page
	 * 
	 * @param testCallerDetails
	 * @param debugs
	 */
	protected static void addDebugInfo(String[] testCallerDetails, List<String> debugs) {
		if (testCallerDetails==null || testCallerDetails.length==0 || debugs==null || debugs.size()==0) {
			return;
		}
	      if (!isDetailReportEnabled()) {
	            return;    //detail reporting not supported
	        }

		String pack = testCallerDetails[0];
		String suite = testCallerDetails[1];
		String testCaseName = testCallerDetails[2];
		if (logInfo) {
	        logger.trace ("pack = " + pack);
	        logger.trace ("suite = " + suite);
	        logger.trace ("testCaseName = " + testCaseName);
		}
		
		
		Package p = getPackage(pack);
		Suite s = getSuite(p, suite);
		TestCase tc = getTestCase(s, testCaseName, parametizedValues);
		
		if (tc == null) {
		    logger.warn ("TestCase is null - " + pack + " :: " + suite + " :: " + testCaseName);
		} else {
		    if (tc.debugs == null || tc.debugs.size () == 0) {
	            tc.debugs = new LinkedList<String>();
		    }
	        for (String str : debugs) {
	            tc.debugs.add(str);
	        }
		}
		
	}
	
	// -------------------------------------------------------------
	// ----- Writing files one test case at a time
	// ----- is invoked by Validator.junitAssertNoErrors()
	// -------------------------------------------------------------

	/**
	 * This implementation is NOT efficient because extra file I/O is done
	 * if default reporting is enabled. It also assumes the Client correctly invoked Validator.junitAssertNoErrors()
	 * 
	 * 1. Find the appropriate TestCase & write the file
	 * 2. Update the co-responding Suite file (may be extra writes)
	 * 3. Update the co-responding Package file (may be extra writes)
	 * 4. Update QAToolsOrg-detail.html
	 * 
	 * 5. Update the co-responding ErrorDetail file (may be extra writes)
	 * 6. Update QAToolsOrg-errorSummary.html
	 */
	protected static void autoGenerateDetailReport(String[] testCallerDetails) {
		if (!isAutoGenerateDetailReportEnabled()) {
		    if (logInfo) {
	            logger.trace("isAutoGenerateDetailReportEnabled - false");
		    }
			return;
		}
		
		//
		// something must be enabled otherwise nothing to do
		if (!isDetailReportEnabled()) {
		    if (logInfo) {
	            logger.trace("isDetailReportEnabled - false");
		    }
			return;
		}
		
		//
		// keep track of number of invokations
		if (numDetailWrites == maxDetailWrites) {
			DetailReport.disableAutoGenerateDetailReport();
			DetailReport.disableDetailReport();
			AdvancedValidator.disableStaticDebugging();
			AdvancedValidator.resetStaticProperties();
		}
		numDetailWrites++;
		
		//
		//
		String pack = testCallerDetails[0];
		String suite = testCallerDetails[1];
		String testcaseName = testCallerDetails[2];
		if (logInfo) {
            logger.trace ("pack = " + pack);
            logger.trace ("suite = " + suite);
            logger.trace ("testcaseName = " + testcaseName);
		}
		
		//
		//
		Package p = getPackage(pack);
		Suite s = getSuite(p, suite);
		List<TestCase> testcases = getTestCases(s, testcaseName);
		
		if (testcases==null || testcases.size ()==0) {
		    logger.warn ("autoGenerateDetailReport - no testcases to report");
		    return;
		}
		
		for (TestCase tc : testcases) {			
			//
			// error summary report
			try {
				for (Detail d : tc.details) {
					ErrorDetail ed = d.errorDetail;
					writeErrorDetail(ed);
				}
				writeErrorSummaryReport();
			} catch (Exception e) {
				logger.warn ("Exception caught in autoGenerateDetailReport - error summary report");
			}

			//
			// this test case
			try {
				writeTestCase(tc);
			} catch (Exception e) {
                logger.warn ("Exception caught in autoGenerateDetailReport = writeTestCase");
			}
		}	//for TestCaes tc
		
		//
		// write suite, package, detail page
		try {
			//write in this order because pass/fail stats are tabulated
			writeSuite(s);
			writePackage(p);
			writeDetailReport();
		} catch (Exception e) {
            logger.warn ("Exception caught in autoGenerateDetailReport = writeSuite, writePackage, writeDetailReport");
		}
		
	}
	
	// -------------------------------------------------------------
	// ----- Writing one file at a time
	// -------------------------------------------------------------

	/**
	 * 
	 */
	private static List<String> getPageHeader(String msg) {
		List<String> list = new LinkedList<String>();
		list.add("<html><head><title>" + msg + "</title></head>");
		list.add("<body><h3>" + msg + "</h3>");
		return list;
	}
	
	/**
	 * //links || QATools.org || Error Summary || Detail Report > package > suite > testcase
	 * 
	 * @return
	 */
	private static String getLinks() {
		StringBuffer sb = new StringBuffer();
		sb.append("<a href=\"QATools.org-errorSummaryReport.html\">Error Summary</a> || ");
		sb.append("<a href=\"QATools.org-detailReport.html\">Detail Report</a>");		
		return sb.toString();
	}

	/**
	 * //links || QATools.org || Error Summary || Detail Report > package > suite > testcase
	 * 
	 * @return
	 */
	private static String getLinks(Package pack, Suite suite, TestCase tc) {
		StringBuffer sb = new StringBuffer();
		sb.append("<a href=\"../QATools.org-errorSummaryReport.html\">Error Summary</a> || ");
		sb.append("<a href=\"../QATools.org-detailReport.html\">Detail Report</a>");
		if (pack != null) {
			sb.append(" &gt; <a href=\"" + pack.htmlFile.getFilename() + "\">" + pack.packageName + "</a>");
		}
		if (suite != null) {
			sb.append(" &gt; <a href=\"" + suite.htmlFile.getFilename() + "\">" + suite.suiteName + "</a>");
		}
		if (tc != null) {
			sb.append(" &gt; <a href=\"" + tc.htmlFile.getFilename() + "\">" + tc.testcaseName + "</a>");
		}
		
		return sb.toString();
	}

	/**
	 * 
	 */
	private static List<String> getPageFooter() {
		List<String> list = new LinkedList<String>();
		list.add("</body>");
		list.add("</html>");
		return list;
	}

	/**
	 * Wraps the Lines in an html <table> 
	 * 
	 * @param lines
	 * @return
	 */
	private static List<String> formatLines(List<Line> lines) {
		if (lines==null || lines.size()==0) {
			return null;
		}
		
		List<String> out = new LinkedList<String>();
		out.add("<table border=2>");
		for (Line l : lines) {
			String color = YELLOW;
			if (l.passFlag == Boolean.TRUE) {
				color = GREEN;
			} else if (l.passFlag == Boolean.FALSE) {
				color = RED;
			}
			
			String tmp = "<tr bgcolor=\"" + color + "\">";
			if (0 < l.num) {
				tmp += "<td>" + l.num + "</td>";
			}
			
			if (Boolean.TRUE.equals(l.passFlag)) {
				tmp += "<td>Pass</td>";
			} else if (Boolean.FALSE.equals(l.passFlag)) {
				tmp += "<td>Fail</td>";
			}
			
			if (l.col1 != null) {
				tmp += "<td>" + l.col1;
				if (l.col1Link != null) {
					tmp += " <a href=\"" + l.col1Link + "\">(more)</a>";
				}
				tmp += "</td>";
			}	//if l.col1

			if (l.col2 != null) {
				tmp += "<td>" + l.col2;
				if (l.col2Link != null) {
					tmp += " <a href=\"" + l.col2Link + "\">(more)</a>";
				}
				tmp += "</td>";
			}	//if l.col2
			
			if (l.col3 != null) {
				tmp += "<td>" + l.col3;
				if (l.col3Link != null) {
					tmp += " <a href=\"" + l.col3Link + "\">(more)</a>";
				}
				tmp += "</td>";
			}	//if l.col3

			
			tmp += "</tr>";
			out.add(tmp);
		}	//for Line
		out.add("</table>");
		return out;
	}
	
	/**
	 * Writes this package to disk
	 */
	private static void writePackage(Package pack) {
		List<String> page = getPageHeader("Package: " + pack.packageName);
		page.add(getLinks(pack,null,null));
		page.add(notEfficientMsg2());
		pack.numWrites++;
		
		pack.total=0;
		pack.fail=0;
		pack.pass=0;
		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<pack.suites.size(); i++) {
			Suite s = pack.suites.get(i);
			Line line = new Line();
			line.num = i+1;
			line.col1 = "Suite: " + s.suiteName;
			line.col1Link = "../../" + s.htmlFile.getFullPath();
			line.col2 = "Total: " + s.total + ", Pass: " + s.pass + ", Fail: " + s.fail;
			
			double pass = s.pass;
			double total = s.total;
			double percent = (pass / total) * 100;
			line.col3 = "" + percent + "%";
			line.passFlag = s.passFlag;	// && !s.assertFlag;
//			if (!s.assertFlag) {
//				line.col3 = ASSERTION_WITHOUT_FOLLUP_CHECK_MSG;
//			}
			lines.add(line);
			
			pack.total += s.total;
			pack.pass += s.pass;
			pack.fail += s.fail;
		}
		page.addAll(formatLines(lines));
		page.addAll(getPageFooter());
		
		try {
			LocalFileUtil.writeLines(pack.htmlFile, page);			
		} catch (Exception e) {
			logger.warn("Exception caught in writePackage");
		}
	}

	/**
	 * Writes this Suite to disk
	 */
	private static void writeSuite(Suite suite) {
		List<String> page = getPageHeader("Package: " + suite.pack.packageName + "<br>Suite: " + suite.suiteName);
		page.add(getLinks(suite.pack,suite,null));
		page.add(notEfficientMsg2());
		suite.numWrites++;

		//
		// it will only be invoked once if the user sets up @AfterClass to invoke the DetailReport
		if (1 < suite.numWrites && notEfficientWarning==null) {
			notEfficientWarning = Boolean.TRUE;
		}

		
		suite.total = suite.testcases.size();
		suite.pass=0;
		suite.fail=0;
		List<Line> lines = new LinkedList<Line>();
		int i;
		for (i=0; i<suite.testcases.size(); i++) {
			TestCase tc = suite.testcases.get(i);
			Line line = new Line();
			line.num = i+1;
			line.col1 = "Testcase: " + tc.testcaseName;
			if (tc.parametizedValues != null) {
				line.col1 += "(" + tc.parametizedValues + ")";
			}
			
			line.col1Link = "../../" + tc.htmlFile.getFullPath();
			line.col2 = "Total: " + tc.total + ", Pass: " + tc.pass + ", Fail: " + tc.fail;
			line.passFlag = tc.passFlag;	// && !tc.assertFlag;
//			if (!tc.assertFlag) {
//				line.col3 = ASSERTION_WITHOUT_FOLLUP_CHECK_MSG;
//			}
			lines.add(line);
			
			if (tc.passFlag) {
				suite.pass++;
			} else {
				suite.fail++;
			}
				
		} // for loop
		
		

		page.addAll(formatLines(lines));
		page.addAll(getPageFooter());
		
		try {
			LocalFileUtil.writeLines(suite.htmlFile, page);
		} catch (Exception e) {
			logger.warn("Exception caught in writeSuite");
		}
	}
	
	/**
	 * Writes this TestCase to disk
	 */
	private static void writeTestCase(TestCase tc) {
		String title = "Package: " + tc.suite.pack.packageName + "<br>Suite: " + tc.suite.suiteName + "<br>TestCase: " + tc.testcaseName;
		if (tc.parametizedValues != null) {
			title += "(" + tc.parametizedValues + ")";
		}
		List<String> page = getPageHeader(title);
		page.add(getLinks(tc.suite.pack,tc.suite,tc));
		page.add(notEfficientMsg2());
		tc.numWrites++;

		//
		// details
		List<Line> lines = new LinkedList<Line>();
		int i;
		for (i=0; i<tc.details.size(); i++) {
			Detail d = tc.details.get(i);
			Line line = new Line();
			line.num = i+1;
			line.col1 = d.msg;
			line.col1Link = null;
			line.passFlag = d.passFlag;
			lines.add(line);
		}
		
//		//
//		// warn if an assertion was not followed up with a junitAssertNoErrors
//		if (!tc.assertFlag) {
//			Line line = new Line();
//			line.num = i + 1;
//			line.col1 = ASSERTION_WITHOUT_FOLLUP_CHECK_MSG;
//			line.col1Link = null;
//			line.passFlag = false;
//			lines.add(line);
//		}
		
		page.addAll(formatLines(lines));

		//
		// debugs
		if (tc.debugs != null && 0<tc.debugs.size()) {
			lines = new LinkedList<Line>();
			for (i=0; i<tc.debugs.size(); i++) {
				Line line = new Line();
				line.num = i+1;
				line.col1 = tc.debugs.get(i);
				line.col1Link = null;
				lines.add(line);
			}
			page.add("<h4>Debug information</h4>");
			page.addAll(formatLines(lines));
		}
		
		page.addAll(getPageFooter());
		
		try {
	        LocalFileUtil.writeLines(tc.htmlFile, page);
		} catch (Exception e) {
		    logger.warn ("Exception caught in writeTestCase");
		}
	}
	
	/**
	 * Writes the main detail page to disk
	 */
	private static void writeDetailReport() {
		List<String> page = getPageHeader("Detail Report");
		page.add(getLinks());
		page.add(notEfficientMsg());

		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<packages.size(); i++) {
			Package p = packages.get(i);
			       
			Line line = new Line();
			line.num = i+1;
			line.col1 = "Package: " + p.packageName;
			line.col1Link = "../" + p.htmlFile.getFullPath();
			line.col2 = "Total: " + p.total + ", Pass: " + p.pass + ", Fail: " + p.fail;
			
			double pass = p.pass;
			double total = p.total;
			double percent = (pass / total) * 100;
			line.col3 = "" + percent + "%";

			line.passFlag = p.passFlag;	// & !p.assertFlag;
//			if (!p.assertFlag) {
//				line.col3 = ASSERTION_WITHOUT_FOLLUP_CHECK_MSG;
//			}
			lines.add(line);
		}
		page.addAll(formatLines(lines));
		page.addAll(getPageFooter());
		LocalFile lf = new LocalFile("QATools.org-detailReport.html");
		
      try {
          LocalFileUtil.writeLines(lf, page);
        } catch (Exception e) {
            logger.warn ("Exception caught in writeDetailReport");
        }

	}

	/**
	 * Writes this error detail to disk
	 */
	private static void writeErrorDetail(ErrorDetail errorDetail) {
		if (errorDetail == null) {
			//System.out.println("writeErrorDetail - null");
			return;
		}
		errorDetail.numWrites++;
		
		List<String> page = getPageHeader("ErrorDetail: " + errorDetail.callingClass + "::" + errorDetail.callingMethod + "::" + errorDetail.callingLine);
		page.add(getLinks(null,null,null));
		page.add(notEfficientMsg2());

		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<errorDetail.details.size(); i++) {
			Detail d = errorDetail.details.get(i);
			String tc = d.testcase.testcaseName;
			if (d.testcase.parametizedValues != null) {
				tc += "(" + d.testcase.parametizedValues + ")";
			}
			String suite = d.testcase.suite.suiteName;
			String pack = d.testcase.suite.pack.packageName;
			
			Line line = new Line();
			line.num = i+1;
			line.col1 = pack + " :: " + suite + " :: " + tc;
			line.col2 = d.msg;
			line.col2Link = "../../" + d.testcase.htmlFile.getFullPath();
			line.passFlag = d.passFlag;
			lines.add(line);
		}
		page.addAll(formatLines(lines));		
		page.addAll(getPageFooter());
		
      try {
          LocalFileUtil.writeLines(errorDetail.htmlFile, page);
        } catch (Exception e) {
            logger.warn ("Exception caught in writeErrorDetail");
        }

	      
	}
	
	/**
	 * Writes the main error summary page to disk
	 */
	private static void writeErrorSummaryReport() {
		List<String> page = getPageHeader("Error Summary Report");
		page.add(getLinks());
		page.add(notEfficientMsg());
		
		Collections.sort(errorDetails);
		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<errorDetails.size(); i++) {
			ErrorDetail ed = errorDetails.get(i);
			if (1 < ed.numWrites && notEfficientWarning==null) {
				notEfficientWarning = Boolean.TRUE;
			}
			String errorCode = (ed.errorCode==null ? "" : " - " + ed.errorCode);
			Line line = new Line();
			line.num = 0;
			line.col1 = ed.details.size() + " instances of ";
			line.col2 = ed.callingClass + "::" + ed.callingMethod + "::" + ed.callingLine + errorCode;
			line.col2Link = "../" + ed.htmlFile.getFullPath();
			line.passFlag = Boolean.FALSE;
			lines.add(line);
		}
		List<String> formattedLines = formatLines(lines);
		if (formattedLines!=null && 0<formattedLines.size()) {
			page.addAll(formattedLines);
		}
		page.addAll(getPageFooter());
		LocalFile lf = new LocalFile("QATools.org-errorSummaryReport.html");
		
		try {
			LocalFileUtil.writeLines(lf, page);
        } catch (Exception e) {
            logger.warn ("Exception caught in writeErrorSummaryReport");
        }	      
	}

	/**
	 * The path to this file is different from the detail & errorSummaryPage than pages in QAToolsOrg directory
	 * 
	 * @return
	 */
	private static String notEfficientMsg() {
		if (Boolean.TRUE.equals (notEfficientWarning)) {
			return "<h5>WARNING - Excessive File I/O - <a href=\"../" + notEfficientFile.getFullPath() + "\">Please Read</a></h5>";
		}
		return "";
	}
	
	/**
	 * 
	 * Extra ../ since these files are in a diff directory
	 */
	private static String notEfficientMsg2() {
        if (Boolean.TRUE.equals (notEfficientWarning)) {
			return "<h5>WARNING - Excessive File I/O - <a href=\"../../" + notEfficientFile.getFullPath() + "\">Please Read</a></h5>";
		}
		return "";
	}


	// -------------------------------------------------------------
	// ----- Generates / writes all files at once
	// ----- This is more efficient than the default : writeTestCaseReports()
	// -------------------------------------------------------------

	/**
	 * This is the efficient way to generate this report. THe client invokes this at the end and reduces the number of file I/O
	 */
	private static void generateErrorSummaryReport() {
		if (errorDetails==null || errorDetails.size ()==0) {
			if (logInfo) {
			    logger.trace("generateErrorSummaryReport - no error details to report");
			}
		    return;
		} 

		try {
			for (ErrorDetail ed : errorDetails) {
				writeErrorDetail(ed);
			}
			writeErrorSummaryReport();
		} catch (Exception e) {
			logger.warn("Exception caught in generateErrorSummaryReport");
		}
	}
	
	/**
	 * This is the efficient way to generate this report. THe client invokes this at the end and reduces the number of file I/O
	 */
	public static void generateDetailReport() {
		
		//
		// keep track of number of invocations
		if (numDetailWrites == maxDetailWrites) {	//nothing happens if max == -1
            logger.warn ("Max writes - " + numDetailWrites);
			DetailReport.disableAutoGenerateDetailReport();
			DetailReport.disableDetailReport();
		}
		numDetailWrites++;
		generateErrorSummaryReport();

		if (packages==null || packages.size ()==0) {
		    logger.warn ("generateDetailReport - no packages to report");
		    return;
		} 
		
		try {
			// write the reports in this order because the stats are tabulated
			for (Package pack : DetailReport.packages) {
				for (Suite s : pack.suites) {
					for (TestCase tc : s.testcases) {
						DetailReport.writeTestCase(tc);
					}
					DetailReport.writeSuite(s);
				}	// for Suite
				DetailReport.writePackage(pack);
			}	// for Package
			DetailReport.writeDetailReport();
		} catch (Exception e) {
			logger.warn ("Exception caught in generateDetailReport");
		}
	}
	
	
	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------

	/**
	 * 
	 * @param testCallerDetails
	 * @param actualCallerDetails
	 * @param errorCode - show code that shows up in the error summary page
	 * @param msg
	 */
	protected static void recordFail(String[] testCallerDetails, String[] actualCallerDetails, String errorCode, String msg) {
		recordResult(false, testCallerDetails, actualCallerDetails, errorCode, msg);
	}

	/**
	 * 
	 * @param testCallerDetails
	 * @param actualCallerDetails
	 * @param msg
	 */
	protected static void recordPass(String[] testCallerDetails, String[] actualCallerDetails, String msg) {
		String errorCode = null;
		recordResult(true, testCallerDetails, actualCallerDetails, errorCode, msg);
	}

	/**
	 * 
	 * @param passFlag
	 * @param testCallerDetails
	 * @param actualCallerDetails
	 * @param errorCode
	 * @param msg
	 */
	protected static void recordResult(boolean passFlag, String[] testCallerDetails, String[] actualCallerDetails, String errorCode, String msg) {
			if (testCallerDetails==null) {
				logger.warn("testCallerDetails is null");
				return;
			} else if (testCallerDetails.length == 0) {
			    logger.warn("testCallerDetails length zero");
			    return;
			} else if (actualCallerDetails == null) {
			    logger.warn("actualCallerDetails is null");
			    return;
			} else if (actualCallerDetails.length == 0) {
			    logger.warn("actualCallerDetails length zero");
			    return;
			}
			
            StackUnwinder.log(testCallerDetails);
            StackUnwinder.log(actualCallerDetails);

			String pack = testCallerDetails[0];
			String suite = testCallerDetails[1];
			String testCaseName = testCallerDetails[2];
	        if (logInfo) {
	                logger.trace ("recordResult - " + passFlag);
	                logger.trace ("pack = " + pack);
	                logger.trace ("suite = " + suite);
	                logger.trace ("testCaseName = " + testCaseName);
	         }
			
			//
			// detail report
			Detail d = addDetail(passFlag, pack, suite, testCaseName, msg);

			//
			// keep track of error summaries
			if (!passFlag && isDetailReportEnabled()) {
				String actualCallingClass = actualCallerDetails[0];
				String actualCallingMethod = actualCallerDetails[1];
				String actualCallingLine = actualCallerDetails[2];
				ErrorDetail ed = getErrorDetail(actualCallingClass, actualCallingMethod, actualCallingLine);
				if (ed == null) {
					ed = new ErrorDetail(actualCallingClass, actualCallingMethod, actualCallingLine);
					ed.errorCode = errorCode;
					//System.out.println("Adding error detail: " + actualCallingClass + ", " + actualCallingMethod + ", " + actualCallingLine);
					errorDetails.add(ed);
				}
				ed.details.add(d);
				d.errorDetail = ed;
				//System.out.println("adding detail to errorDetail :: " + actualCallingClass + ", " + actualCallingMethod + ", " + actualCallingLine);
			}
		}

	/**
	 * 
	 * @param passFlag
	 * @param pack
	 * @param suite
	 * @param testcaseName
	 * @param detail
	 */
	protected static Detail addDetail(boolean passFlag, String pack, String suite, String testcaseName, String detail) {
		Detail d = new Detail(passFlag, detail);
		if (!isDetailReportEnabled()) {
		    if (logInfo) {
	            logger.trace ("detailReportFlag = " + detailReportFlag);
	            logger.trace ("addDetail - was not added. pack=" + pack + ", suite=" + suite + ", testcaseName=" + testcaseName + ", detail=" + detail);
		    }
			return d;	// return without adding detail to list
		}

		Package p = getPackage(pack);
		if (p == null) {
		    logger.trace ("Creating new package - " + pack);
			p = new Package(pack);
			packages.add(p);
			//System.out.println("Adding package: " + pack);
		}
		
		Suite s = getSuite(p, suite);
		if (s == null) {
            logger.trace ("Creating new suite - " + suite);
			s = new Suite(suite);
			s.pack = p;
			p.suites.add(s);
			//System.out.println("Adding suite: " + suite);
		}
		
		TestCase tc = getTestCase(s, testcaseName, parametizedValues);
		if (tc == null) {
            logger.trace ("Creating new testcase - " + testcaseName);
			tc = new TestCase(testcaseName);
			tc.suite = s;
			s.testcases.add(tc);
			if (parametizedValues != null) {
				tc.parametizedValues = new String (parametizedValues);
			}
//			System.out.println("Adding testcase: " + testcaseName);
		}
		
		//
		//detail report shows an error if client does NOT check for errors
//		p.assertFlag = true;
//		s.assertFlag = true;
//		tc.assertFlag = true;	
		
		
		d.testcase = tc;
		tc.details.add(d);
//		System.out.println("Adding detail: " + detail);
		tc.total++;		
		if (!passFlag) {
			p.passFlag = false;
			s.passFlag = false;
			tc.passFlag = false;
			tc.fail++;
		} else {
			tc.pass++;
		}
		return d;
	}
	
	/**
	 * Validator resets the flag so the DetailReport doesn't flag an issue.
	 * This warns if an assertion is invoked without a followup junitAssertNoErrors() call
	 */
	
	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------
	
	/**
	 * 
	 * @param pack
	 * @return
	 */
	protected static Package getPackage(String pack) {
		if (pack==null || pack.length()==0 || packages==null || packages.size()==0) {
			return null;
		}
		Package p = null;
		for (Package pp : packages) {
			if (pp.packageName.equals(pack)) {
				p = pp;
				break;
			}
		}
		
		if (p == null) {
		    logger.trace ("Package not found - " + pack);
		} else {
            logger.trace ("Package found - " + pack);
		}
		return p;
	}

	/**
	 * 
	 * @param pack
	 * @param suite
	 * @return
	 */
	protected static Suite getSuite(Package pack, String suite) {
		if (pack==null || pack.packageName.length()==0 || suite==null || suite.length()==0 || pack.suites==null || pack.suites.size()==0) {
			return null;
		}

		Suite p = null;
		for (Suite pp : pack.suites) {
			if (pp.suiteName.equals(suite)) {
				p = pp;
				break;
			}
		}
        if (p == null) {
            logger.trace ("Suite not found - " + pack + " :: " + suite);
        } else {
            logger.trace ("Suite found - " + pack + " :: " + suite);
        }
		return p;
	}

	/**
	 * 
	 * @param suite
	 * @param testcaseName
	 * @param parametizedValues
	 * @return
	 */
	protected static TestCase getTestCase(Suite suite, String testcaseName, String parametizedValues) {
		if (suite==null || suite.testcases.size()==0 || testcaseName==null || testcaseName.length()==0) {
			return null;
		}
		TestCase p = null;
		for (TestCase pp : suite.testcases) {
			if (pp.testcaseName.equals(testcaseName)) {
				String a = "" + pp.parametizedValues; //add blank so it's easy to compare nulls. parametizedValues can be null
				String b = "" + parametizedValues;
				if (a.equals(b)) {
					p = pp;
					//System.out.println("Found testcase: " + testcaseName);
					break;
				}
			}
		}
        if (p == null) {
            logger.trace ("TestCase not found - " + suite.pack.packageName + " :: " + suite + " :: " + testcaseName);
        } else {
            logger.trace ("TestCase found - " + suite.pack.packageName + " :: " + suite + " :: " + testcaseName);
        }
		return p;
	}

	/**
	 * 
	 * @param suite
	 * @param testcaseName
	 * @return
	 */
	protected static List<TestCase> getTestCases(Suite suite, String testcaseName) {	
		if (suite==null || suite.testcases.size()==0 || testcaseName==null || testcaseName.length()==0) {
			return null;
		}
		List<TestCase> p = new LinkedList<TestCase>();
		for (TestCase pp : suite.testcases) {
			if (pp.testcaseName.equals(testcaseName)) {
					p.add(pp);
					//System.out.println("Found testcase: " + testcaseName);
			}
		}
		if (p.size()==0) {
			return null;
		}
		return p;
	}

	
	/**
	 * 
	 * @param actualCallingClass
	 * @param actualCallingMethod
	 * @param actualCallingLine
	 * @return
	 */
	protected static ErrorDetail getErrorDetail(String actualCallingClass, String actualCallingMethod, String actualCallingLine) {
		if (actualCallingClass==null || actualCallingClass.length()==0 || actualCallingMethod==null || actualCallingMethod.length()==0 || actualCallingLine==null || actualCallingLine.length()==0) {
			return null;
		}
		ErrorDetail ed = null;
		for (ErrorDetail e : errorDetails) {
			if (e.callingClass.equals(actualCallingClass)
					&& e.callingMethod.equals(actualCallingMethod)
					&& e.callingLine .equals(actualCallingLine)
					) {
				ed = e;
				break;
			}
		}
		return ed;
	}

	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------

	
	/**
	 * 
	 * @author hockeyeh
	 *
	 */
	private static class Package {
		public String packageName;
		public List<Suite> suites = new LinkedList<Suite>();
		public Boolean passFlag = Boolean.TRUE;
		public int total=0;
		public int pass=0;
		public int fail=0;
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "p-" + LongUtil.getUniqueLong2() + ".html");
		public int numWrites = 0;
		//public boolean assertFlag = false; //keep track when assertions are invoked without invoking junitAssertNoErrors
		public Package(String name) { packageName = name; }
		
	}
	
	private static class Suite {
		public String suiteName;
		public List<TestCase> testcases = new LinkedList<TestCase>();
		public Boolean passFlag = Boolean.TRUE;
		public int total=0;
		public int pass=0;
		public int fail=0;
		public Package pack;
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "s-" + LongUtil.getUniqueLong2() + ".html");
		public int numWrites = 0;
		//public boolean assertFlag = false; //keep track when assertions are invoked without invoking junitAssertNoErrors
		public Suite(String name) { suiteName = name; }
	}
	
	private static class TestCase {
		public String testcaseName;
		public List<Detail> details = new LinkedList<Detail>();
		public Boolean passFlag = Boolean.TRUE;
		public int total=0;
		public int pass=0;
		public int fail=0;
		public Suite suite;
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "tc-" + LongUtil.getUniqueLong2() + ".html");
		public int numWrites = 0;
		public List<String> debugs;
		public String parametizedValues = null;
		//public boolean assertFlag = false; //keep track when assertions are invoked without invoking junitAssertNoErrors
		public TestCase(String name) { testcaseName = name; }
	}
	
	private static class Detail {
		public String msg;
		public Boolean passFlag = Boolean.TRUE;
		public TestCase testcase;
		public ErrorDetail errorDetail;
		public Detail(boolean passFlag, String detail) {msg = detail; this.passFlag = passFlag; }
	}
	protected static class ErrorDetail implements Comparable {
		public String callingClass;
		public String callingMethod;
		public String callingLine;
		public String errorCode;	//this shows up on the error summary page
		public List<Detail> details = new LinkedList<Detail>();
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "ed-" + LongUtil.getUniqueLong2() + ".html");
		public int numWrites = 0;
		public ErrorDetail(String callingClass, String callingMethod, String callingLine) { this.callingClass = callingClass; this.callingMethod = callingMethod; this.callingLine = callingLine; }
		
		public int compareTo(Object arg0) {
			return ((ErrorDetail) arg0).details.size() - this.details.size();
		}
	}
	
	private static class Line {
		public Boolean passFlag = null;	//null = Yellow Debug
		public int num;
		public String col1 = null;
		public String col2 = null;
		public String col3 = null;
		public String col1Link = null;
		public String col2Link = null;		
		public String col3Link = null;		
	}
	
}



